//  Model railway turntable control program, using a micro-stepping stepper motor
//  Version for use with  Pololu MP6500 microstepping controller
//  Assumes 400-step motor with 1/8 microsteps

/*   GENERAL NOTES   
 *    Sketch is 'blocking', which is OK for this application, and probably makes it simpler to follow.
 *    
 *    Motion is controlled with either:
 *    - Pushbuttons connected to digital input pins on the arduino.  This will be the normal way of operating.
 *    - An infra-red key pad.  It duplicates most of the button functions, and has a few maintenance routines as well.
 *    
 *    Key routines are
 *    -  doStep().  Issues step and direction commands to driver.
 *    -  doMove().  Commands moves, keeps track of current TT location, and allows for motion cancellation.  Calls doStep().
 *    
 *    Stepping speed is controlled by variable 'waitTime'.
 *    
 *    Direction of rotation is normally manually controlled, using variable 'CW'.  User has a button control, and LEDs 
 *    show the current direction.
 *    
 *    Th eventual intent is to set and define a 'zero' position with the 'OK' key, then remember it using EEPROM
 *    during power OFF times.  There is no hardware 'key' to set and maintain this.  I have yet to see if this works OK.
 *    Obviously it will rely on accurate stepping motion, and no power on/off jumping by the motor.
 *    
 *    If it should prove that I need to implement the motion offset, it should be noted that it won't be cumulative.  At any
 *    time there will be no more than 'offset' microsteps seperation from the nominal position (one way or the other).
 */
/*
  Connections
  stepPin on Arduino (10) to Step on driver
  dirPin on Arduino (11) to Dir on driver
  holdPin on Arduino (5) to i2 on driver
  enablePin (6) on Arduino to ENBL on driver
  VIN on Arduino to +5v regulator IC
  IR S to Pin 0 on Arduino
  IR - to GND on Arduino
  IR VCC to 5V on Arduino
  MS1, MS2, SLEEP pins on driver to +5v regulator
  Pushbuttons to pins on Arduino
    9 = star (turn TT round)
    A0 = hash (change direction of rotation)
    0 = track 0 (entry track)
    1, 3, 4, 7, 8 other tracks
    Could use analogue pins if more tracks needed
    2 = cancel motion interrupt pin
  Pin 13 on Arduino to direction indicator LEDs
*/
#include <IRLib.h>
#include <Wire.h>
#include <EEPROM.h>

int RECV_PIN = 12;

IRrecv My_Receiver(RECV_PIN);
IRdecode My_Decoder;
IRdecodeHash My_Hash_Decoder;

//  These are identified empirically, to suit the particular keypad
#define RIGHT_ARROW   16734885
#define LEFT_ARROW    16716015
#define OK_BUTTON     16726215
#define UP_ARROW      16718055
#define DOWN_ARROW    16730805
#define BUTTON_0      16750695
#define BUTTON_1      16753245
#define BUTTON_2      16736925
#define BUTTON_3      16769565
#define BUTTON_4      16720605
#define BUTTON_5      16712445
#define BUTTON_6      16761405
#define BUTTON_7      16769055
#define BUTTON_8      16754775
#define BUTTON_9      16748655
#define STAR          16738455
#define HASHKEY       16756815

int stepPin = 10;
int dirPin = 11;
int holdPin = 5;
int enablePin = 6;

int BTN0 = 0;
int BTN1 = 1;
int BTN2 = 3;
int BTN3 = 4;
int BTN4 = 7;
int BTN5 = 8;
int BTN_turn = 9;
int PINLED = 13;
int BTN_direction = 14;  //analog pin A0

const int defaultOffset = 3;    //  Extra microsteps on reversal
const int minOffset = 0;
const int maxOffset = 10;
int offset = 0;    // will probably set to 3 or 4 if I eventually decide to go this route.

boolean lastDir = true;   // previous direction of rotation

unsigned long waitTime =10;  // Pulse half-time in mS.  Initial value = 10 mS - approx 1 rpm
const int fullTurn = 100;  // number of available track positions around the turntable
const int halfTurn = fullTurn / 2;
boolean CW = true;   // direction of rotation
boolean moveTT;   // Whether or not to do a TT move
int destination = 0;   // Desired end position
int location = 0;   // TT bridge actual position
long unsigned mvalue ;  // Value returned after keypad operation
const unsigned long debounceDelay = 100;    // the debounce time in mS; increase if CW reversal isn't consistent
const unsigned long noiseFilterDelay = 2;   //  will screen out noise on button inputs less than this value (mS)
const int right = true;
const int left = false;
const boolean highPower = true;
const boolean lowPower = false;
boolean lastPower = highPower;   // ensure motorPower() will run during setup()

volatile boolean cancel = false;  // when we want to cancel the current motion

//Position of tracks on turntable
const int  track0 = 0 ;
const int  track1 = 28 ;
const int  track2 = 56 ;
const int  track3 = 64 ;
const int  track4 = 74 ;
const int  track5 = 92 ;


void setup() {

  pinMode(stepPin, OUTPUT);
  pinMode(dirPin, OUTPUT);
  pinMode(holdPin, INPUT);   //set to high impedance state, for low power operation
  pinMode(enablePin, OUTPUT);  //   Not used in this sketch

  pinMode(PINLED, OUTPUT);
  digitalWrite(PINLED, HIGH);   //Initialize direction LEDs

  pinMode(BTN0, INPUT_PULLUP);
  pinMode(BTN1, INPUT_PULLUP);
  pinMode(BTN2, INPUT_PULLUP);
  pinMode(BTN3, INPUT_PULLUP);
  pinMode(BTN4, INPUT_PULLUP);
  pinMode(BTN5, INPUT_PULLUP);
  pinMode(BTN_turn, INPUT_PULLUP);
  pinMode(BTN_direction, INPUT_PULLUP);
  pinMode(2, INPUT_PULLUP);  //  Interrupt pin

  digitalWrite (stepPin, LOW);
  digitalWrite (dirPin, LOW);
  digitalWrite (enablePin, LOW);
  motorPower (lowPower);

  My_Receiver.enableIRIn(); // Start the receiver
  
  Serial.begin(9600);    //just in case we want to write anything back over USB.  Testing use only.

  attachInterrupt (digitalPinToInterrupt (2), stopMotion, FALLING);

  //location = (int) EEPROM.read(0);  //get stored value of TT location
  //if (location >= fullTurn || location < 0) location = 0;  //first time executed only - normally read location will be used
}

// respond to interrupts on pin D2.  For stopping motion early.  Stop will only occur at the next 32-microstep boundary.
void stopMotion() {
  cancel = true;
}

void motorPower(boolean power)
{
  if (power == highPower && lastPower == lowPower) {
    pinMode(holdPin, OUTPUT);
    digitalWrite ( holdPin, LOW);   // set motor current to 1.5A
    lastPower = highPower;
  }
  else if (power == lowPower && lastPower == highPower) {
    pinMode (holdPin, INPUT);  // set motor current to 0.5A
    lastPower = lowPower;
  }
}

void changeDirection()
{
  CW = !CW;
  if ( CW ) digitalWrite(PINLED, HIGH);
  else digitalWrite(PINLED, LOW);
}

//  Filters out short noise spikes on button inputs (if any)
//  Requires input to be LOW for at least 'noiseFilterDelay' before being recognised
int readButton(int button)
  {
  unsigned long endTime;
  endTime = millis() + noiseFilterDelay;  // when to stop checking input
  do {
    if (digitalRead (button)  ==  HIGH) return HIGH; //  nothing's happening or noise present
  }
  while (millis() <= endTime);    // keep on reading button
  return LOW;  //returns LOW if button stays low for delay duration
}

void blinkLEDS()    //  acknowledge position reset command
{
  if (digitalRead (PINLED) == HIGH)
  {
    digitalWrite (PINLED, LOW);
    delay (500);
    digitalWrite (PINLED, HIGH);
  }
  else
  {
    digitalWrite (PINLED, HIGH);
    delay (500);
    digitalWrite (PINLED, LOW);
  }
  delay (500);
}

void doStep (int steps, boolean rotation)    // steps the motor between successive 'home states'
  {
    if (steps <= 0) return;
    int microsteps = steps * 32;
    if (digitalRead (enablePin) == HIGH) {
     digitalWrite (enablePin, LOW); // wake up driver if not already awake
     delay(1);
    }
    motorPower (highPower);    // set for full 1.5A current
    if (rotation) digitalWrite (dirPin, HIGH);
         else digitalWrite (dirPin, LOW);
    if (rotation != lastDir) 
    {
      lastDir = rotation;
      microsteps += offset;  // offset to allow for system hysteresis (or whatever)
      delay (1);   //  allow time for driver to settle after change of direction
    }  
    for (int i = 0; i < microsteps; i++)
    {
      digitalWrite (stepPin, HIGH);
      delay (waitTime);
      digitalWrite (stepPin, LOW);
      delay (waitTime);
    }
  }

//  Actual TT move routine
void doMove(int distance )
{
  cancel = false;  // clear any previous cancel requests
  for ( int i = 0; i < distance; i++)
  {
    doStep(1, CW);  // step one at a time to keep track of where we are
    if (CW) location = (location + 1) % fullTurn;
    else
    {
      location--;
      if (location < 0) location += fullTurn;
    }
    if (cancel) break; //if interrupt (cancel) has occurred during previous motion, end all motion
  }
}


// the loop routine runs over and over again forever:
void loop()
{
  /*  Keypad commands:
       0 - track 0 (entry track)
       1, 2, 3, 4, 5 - Storage tracks  (could add more if required).
       OK - Register current TT position as track 0.
       Left, Right Arrow - One increment left or right.  50 increments make a full circle.
       Up arrow - increase offset by 1    )  Blink direction LEDS
       Down arrow - reduce offset by 1    )     to confirm
       Star - rotate 180 degrees - i.e. turn turntable round from any starting position
       Hash - Alternate clockwise or anti-clockwise direction of rotation.  LED to indicate current selection.
       Note that TT will never move more than 180 degrees, since closest end of TT bridge to the commanded destination
         will stop when it gets there.  User must select appropriate direction of rotation to ensure that the engine
         will come off the TT bridge facing the required way.
       Selecting the track that the TT is currently at will do nothing.  To turn a half circle requires the * key.

      Button commands to input pins:
      0, 1, 3, 4, 7, 8 move to track 0, 1, 2, 3, 4, 5
      2 - cancel current motion
      9 - as Star - half turn
      A0 (D14) - as hash - change direction
      Note that keypad OK and L, R, U, D arrows do not have a button equivalent, since they won't be used in normal operation
      Also, 'cancel' function not available on keypad (button feature only), but it will cancel a keypad motion

      Each 'step' in doStep() actually corresponds to 4 stepper-motor steps, and is intended to move the motor between
         successive 'home states' of the MP6500.  As it is a 400-step motor, 100 application-'steps' are required for a full rotation.
         8 microsteps per full motor step is assumed, therefore 32 microsteps are required for each application-'step'.
         'location' keeps track of where we are, and an application-'step' of 1 increments/decrements 'location' by 1.
  */

  mvalue = 0;
  moveTT = false;

  if ( CW ) digitalWrite(PINLED, HIGH);
  else digitalWrite(PINLED, LOW);

  if (My_Receiver.GetResults(&My_Decoder))
  { //Puts results in My_Decoder
    My_Hash_Decoder.copyBuf(&My_Decoder);  //copy the results to the hash decoder
    My_Decoder.decode();
    mvalue = My_Decoder.value;
    My_Receiver.resume();
  }

  if (mvalue == OK_BUTTON)
  {
    location = track0 ;
    offset = defaultOffset;
    blinkLEDS();
  }

//     *******************  I've done most of my testing with these next two operations.  They aren't currently available as  ****
//      buttons, but could temporarily be so assigned if you wanted to run the sketch.
//                          
  else if (mvalue == RIGHT_ARROW )
  {
    doStep(1, right);
    location = (location + 1) % fullTurn;
  }

  else if (mvalue == LEFT_ARROW  )
  {
    doStep(1, left);
    location-- ;
    if (location < 0) location = location + fullTurn;
    //  ***************************************************************************************************
  }
/*  **************  This was so that I can adjust the offset on the fly  *************
 else if (mvalue == UP_ARROW && offset < maxOffset)   //  Increase correction offset
  {
      offset++;
      blinkLEDS();
  }

  else if (mvalue == DOWN_ARROW && offset > minOffset)    //  Reduce correction offset
  {
    offset--;
    blinkLEDS();
  }
  ***************************************************************************/
 
  else if (mvalue == HASHKEY )  changeDirection () ;

  else if ( readButton(BTN_direction) == LOW)
  { // button to change direction
    changeDirection();   // Change direction of turntable rotation
    //  Wait for all Direction button activity to cease before next loop iteration
    delay(debounceDelay);   // delay after pressing button to allow for contact bounce
    while (digitalRead(BTN_direction) == LOW);   // wait for button to be released
    delay(debounceDelay);   // delay after releasing button to allow for contact bounce
  }

     //  ***********  this reads buttons for normal moves, calculates the move distance, and calls the move routine  ******** 
     //               Not essential to see the problem (use RIGHT_ARROW and LEFT_ARROW for this)
     
  else if (mvalue == STAR || readButton(BTN_turn) == LOW)  doMove ( halfTurn );

  else
  {
    moveTT = true;
    if (mvalue == BUTTON_0 || readButton(BTN0) == LOW) destination = track0;
    else if (mvalue == BUTTON_1 || readButton(BTN1) == LOW) destination = track1;
    else if (mvalue == BUTTON_2 || readButton(BTN2) == LOW) destination = track2;
    else if (mvalue == BUTTON_3 || readButton(BTN3) == LOW) destination = track3;
    else if (mvalue == BUTTON_4 || readButton(BTN4) == LOW) destination = track4;
    else if (mvalue == BUTTON_5 || readButton(BTN5) == LOW) destination = track5;
    else moveTT = false;    // All other buttons do nothing

    if (moveTT)
    { // normal operational movements of turntable
      int distToMove;
      if (CW) distToMove = (fullTurn + destination - location) % halfTurn;
      else distToMove = (fullTurn + location - destination) % halfTurn;
      doMove( distToMove );
    }
  }
//     *********************************************************************************************************
  //EEPROM.update(0, (byte) location);  //remember in case of power-off.  Only changes if 'location' has changed.

 //Serial.print(location, DEC);  Serial.print("    ");   Serial.println( pps, DEC);
   
motorPower (lowPower);   // Power down output to 0.5 A - holding current
}
